/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : chromo.cc                                                */
/*                                                                       */
/*       Date : November  6, 1997                                        */
/* Updated at : November 24, 1997                                        */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include "chromo.h"
#include "gene.h"
#include "EdgeGraph.h"
#include "global.h"
#include <stdlib.h>


/*
// Modified random generator.
*/

unsigned int myRand(type_Chromosome* this)
{
   int ret = ((rand() % (this->chromoLength - 2)) + 1);
   return ret;

}// end of myRand() 

/*
//
*/

type_Chromosome* defineChromosome(unsigned int _chromoLength, 
                                  type_Gene *_genePool[])
{
   type_Chromosome* myChro = (type_Chromosome*) calloc(1,sizeof(type_Chromosome));

   if(myChro!=NULL)
   {
      myChro->myGenePool = (type_Gene *) calloc(_chromoLength,sizeof(type_Gene));

      if(myChro->myGenePool!=NULL)
      {
         myChro->chromoLength = _chromoLength;
         myChro->genePool = _genePool;
         myChro->myEval = -1.0;
         myChro->myModEval = -1.0;
      }
      else
        {
            free(myChro);
            myChro = NULL;
        }

   }

   return myChro;

}// end of 

/* Deep copy */

type_Chromosome* copyChromosome(type_Chromosome* RHS) 
{
   int i;
   type_Chromosome* ret = defineChromosome(RHS->chromoLength,RHS->genePool);

   if(ret!=NULL)
   {
      ret->myEval = RHS->myEval;
      ret->myModEval = RHS->myModEval;

      for(i=0;i<RHS->chromoLength;i++) 
          ret->myGenePool[i] = RHS->myGenePool[i];
   }

   return ret;

}

/*
// Chromosome initialization.
*/

void init(type_Chromosome* this, type_Population *myPop)
{
   int i, j, k, notExist=0;

   int chromoLength = myPop->chromoLength;
   char *mark = (char*) calloc(chromoLength,sizeof(char));
   
   for (i = 1; i < (chromoLength - 1); i++) mark[i] = 0;

   mark[0]=mark[chromoLength - 1] = 1;

   this->myGenePool[0] = *myPop->genePool[0];
   this->myGenePool[chromoLength - 1] = *myPop->genePool[chromoLength - 1];
   
   for (i = 1; i < (chromoLength - 1); i++)
   {
      //while (myGenePool[(j = myRand())] != (Gene *)0)
      while (mark[(j = myRand(this))] != 0)
	 continue;
      this->myGenePool[j] = *myPop->genePool[i];
      mark[j] = 1;

   }// end of for

   free(mark);

}// end of init()

/*
// Calculates (if necessary) and returns chromosome evaluation.
*/

eval_t eval(type_Chromosome* this)
{
   int i;
   if (this->myEval == -1.0)
   {
      this->myEval = 0.0;
      for (i = 1; i < (this->chromoLength - 1); i++)
	 if(getIndex(this->myGenePool[i-1])>-1 && 
            getIndex(this->myGenePool[i])>-1 )
	      this->myEval += distanceTo(&this->myGenePool[i-1],&this->myGenePool[i]);

       this->myEval += distanceTo(&this->myGenePool[i],&this->myGenePool[0]); 

       if(!(this->myEval>0) ) this->myEval = -1;

   }// end of if

   return this->myEval;

}// end of eval()

/*
// Keeping track of used genes during crossover.
*/

/*
int isUsed(type_Gene gene, type_Gene *usedGene[], int noOfUsedGenes)
{
   int i, ret = 0;

   for (i = 0; !ret && i < noOfUsedGenes; i++)
      ret = (isEqual(usedGene[i],gene)) ;

   return ret;

}// end of isUsed()
*/

/*
// Keeping track of used genes during crossover.
*/

int isUsed2(type_Gene gene, type_Gene *usedGene, int noOfUsedGenes)
{
   int i, ret = 0;
   
   for (i = 0; !ret && i < noOfUsedGenes; i++)
      ret = isEqual(usedGene[i], gene) ;

   return ret;

}// end of isUsed2()

/*
// Performs crossover using parent1 and parent2 as templates.
// This chromosome is assigned the result.
*/

void crossover(type_Chromosome* this, enum cross_t crossoverType, 
               type_Chromosome* parent1, type_Chromosome* parent2)
{
   int i, j, k, l;
   int curGene = 0;
   int chromoLength = this->chromoLength;
   type_Gene *usedGene = (type_Gene*) calloc(chromoLength, sizeof(type_Gene));
   type_Gene *resGene = (type_Gene*) calloc(chromoLength, sizeof(type_Gene));
   type_Gene *notCommon = (type_Gene*) calloc(chromoLength, sizeof(type_Gene));
   char *mark = (char *) calloc(chromoLength, sizeof(char));
   int  size=0;

   mark[0]=1;
   usedGene[0] = resGene[0] = parent1->myGenePool[0];

   mark[chromoLength - 1]=1;
   usedGene[1] = resGene[chromoLength - 1] = parent1->myGenePool[chromoLength - 1];

   if (crossoverType == fixedPos)
   {
      j = 2;
      for (i = 1; i < (chromoLength - 1); i++)
      {
         mark[i] = 0;
	 if (rand() % 2)
	 {
            mark[i] = 1;
	    resGene[i] = usedGene[j] = parent1->myGenePool[i];
	    j++;
	 }

      }// end of for

   }
   else // crossoverType == fixedPosSeq
   {
      int left, right;
      type_Gene *tmpGene;
      
      left = ((rand() % (chromoLength - 3)) + 1);
      right = ((rand() % (chromoLength - (2 + left))) + left + 1);
      
      j = 2;
      for (i = 1; i < (chromoLength - 1); i++)
      {
         mark[i] = 0;
	 if (i >= left && i <= right)
	 {
            mark[i] = 1;
	    usedGene[j] =  resGene[i] = parent1->myGenePool[i];
	    j++;
	 }

      }// end of for

   }
   
   for(i=0;i<chromoLength;i++)
       if (!(isUsed2(parent2->myGenePool[i], usedGene, j)))
       {  notCommon[size]= parent2->myGenePool[i];
          size++;
       }

   k=l=0;
   while(k<chromoLength)
   {
      if( mark[k] == 0)
      {
	 if(l<size)
	 {
	    resGene[k] = notCommon[l];
	    l++;

	 }// end of if

      }// end of if

      k++;

   }// end of while

   for(k=0;k<chromoLength;k++) this->myGenePool[k] = resGene[k];

   free(notCommon);
   free(resGene);
   free(usedGene);
   free(mark);

   this->myEval = -1.0;
   eval(this);

}// end of crossover()

/*
// Performs mutation on this chromosome.
*/

void mutate(type_Chromosome *this)
{
   int left, right, i, s;
   type_Gene tmpGene;
   int chromoLength = this->chromoLength; 
   
   left = ((rand() % (chromoLength - 3)) + 1);
   right = ((rand() % (chromoLength - (2 + left))) + left + 1);
   
   for (i = left; i <= right; i++)
   {
      tmpGene = this->myGenePool[i];
      s = left + rand() % (right - left);
      this->myGenePool[i] = this->myGenePool[s];
      this->myGenePool[s] = tmpGene;

   }// end of for

   this->myEval = -1.0;

   eval(this);

}// end of mutate()

/*
// Mates this chromosome with partner resulting in two new offsprings, 
// child1 and child2.
*/

void mate(type_Chromosome* this, type_Chromosome* partner,
          float _crossover, enum cross_t crossoverType, float _mutation, 
          type_Chromosome *child1, type_Chromosome *child2)
{

   float rnd = (rand() * 0.00001);
   rnd = rnd - (int)rnd;

   if (rnd <= (_crossover ))
      crossover(child1,crossoverType, this, partner);
   else
      child1 = copyChromosome(this);

   if (rnd <= (_mutation ))
      mutate(child1);
   
   if (rnd <= (_crossover ))
      crossover(child2,crossoverType, partner, this);
   else
      child2 = copyChromosome(partner);
    
   if (rnd <= (_mutation )) mutate(child2);

}// end of mate()

int printChromosome(type_Chromosome * this)
{
   int i = 0;

   for(i=0;i<this->chromoLength;i++) 
       printf(" %2d ",getIndex(this->myGenePool[i]));
   
   printf(" - %f ", eval(this));

}

eval_t setModEval(type_Chromosome* this, eval_t _modEval)
       {return (this->myModEval = _modEval);}

eval_t getModEval(type_Chromosome* this) {return this->myModEval;}

eval_t getEval(type_Chromosome* this) {return this->myEval;}

unsigned int getLength(type_Chromosome* this) {return this->chromoLength;}

void disposeChromosome(type_Chromosome* this) 
{ 
   free(this->myGenePool);
   free(this);
}
